// SingleList.h

#pragma once

#ifndef _SINGLE_LIST_H_
#define _SINGLE_LIST_H_

#include "allocator2.h"
#include "common_macros.h"
#include "HashString.h"


#include <xutility>

//------------------------------------------------------------------------------




//------------------------------------------------------------------------------

using namespace HashString;

class format_ofstream;

//----------------------------------------------------------
// SingleList1TemplateAllocator
//----------------------------------------------------------


//struct single_list1_template_size
//{
//	byte bytes[8];
//};

//typedef FixedSmallObjAllocator
//< single_list1_template_size, 4, OwnPageAllocator >
//	SingleList1TemplateAllocator;

typedef FixedSmallObjAllocator
	< ALIGN_TO_2_POW_N_OR_64 ( 8 ), DefaultAllocatorId, OwnPageAllocator >
		SingleList1TemplateAllocator;


//------------------------------------------------------------------------------
// single_list1
//
// Single-linked list for pointers, T must be pointer.
// List containg at least one element, but in some cases this
// element maybe null.
// This list does not automatically delete nodes,
// for do this use delete_nodes() function.
// Special delete_nodes_and_values() function delete single-list
// nodes and values on which node pointers refers.
//------------------------------------------------------------------------------

template < typename T >
struct single_list1
{
	struct node
	{
		node()
			: value_(T())
			, next_(0)
		{}

		node ( T value )
			: value_(value)
			, next_(0)
		{}

		node ( T value, node * next )
			: value_(value)
			, next_(next)
		{}


		inline void * operator new ( size_t size )
		{
			return SingleList1TemplateAllocator::allocate();
		}

		inline void operator delete ( void * address )
		{
			SingleList1TemplateAllocator::deallocate ( address );
		}


		T value_;
		node * next_;
	};

	class iterator
	{
	public:
		//template<class _Ty>
		//struct iterator_traits//<_Ty *>
		//{	// get traits from pointer
		typedef std::forward_iterator_tag iterator_category;
		typedef T value_type;
		typedef ptrdiff_t difference_type;
		typedef ptrdiff_t distance_type;	// retained
		typedef T * pointer;
		typedef T & reference;
		//};

	public:
		iterator ( node * current, node ** prev )
			: current_ ( current )
			, prev_ ( prev )
		{}

		iterator()
			: current_ ( 0 )
			, prev_ ( 0 )
		{}

		T & operator*() const
		{
			return current_->value_;
		}

		T * operator->() const
		{
			return & current_->value_;
		}

		bool operator== ( const iterator & rhs ) const
		{
			return current_ == rhs.current_;
		}

		bool operator!= ( const iterator & rhs ) const
		{
			return current_ != rhs.current_;
		}

		iterator & operator++ () // preincrement
		{
			prev_ = & current_;
			current_ = current_->next_;
			return *this;
		}

		iterator operator++ ( int ) // postincrement
		{
			iterator retValue = *this;
			prev_ = & current_;
			current_ = current_->next_;
			return retValue;
		}

	private:
		node * current_;
		node ** prev_;
	};

	class const_iterator
	{
	public:
		//template<class _Ty>
		//struct iterator_traits//<_Ty *>
		//{	// get traits from pointer
		typedef std::forward_iterator_tag iterator_category;
		typedef const T value_type;
		typedef ptrdiff_t difference_type;
		typedef ptrdiff_t distance_type;	// retained
		typedef const T * pointer;
		typedef const T & reference;
		//};

	public:
		const_iterator ( const node * current, const node ** prev )
			: current_ ( current )
			, prev_ ( prev )
		{}

		const_iterator()
			: current_ ( 0 )
			, prev_ ( 0 )
		{}

		const T & operator*() const
		{
			return current_->value_;
		}

		const T * operator->() const
		{
			return & current_->value_;
		}

		bool operator== ( const const_iterator & rhs ) const
		{
			return current_ == rhs.current_;
		}

		bool operator!= ( const const_iterator & rhs ) const
		{
			return current_ != rhs.current_;
		}

		const_iterator & operator++ () // preincrement
		{
			prev_ = & current_;
			current_ = current_->next_;
			return *this;
		}

		const_iterator operator++ ( int ) // postincrement
		{
			const_iterator retValue = *this;
			prev_ = & current_;
			current_ = current_->next_;
			return retValue;
		}

	private:
		const node * current_;
		const node ** prev_;
	};

	iterator begin()
	{
		return iterator ( & first_, 0 );
	}

	iterator end()
	{
		return iterator ( 0, & last_ );
	}

	const_iterator begin() const
	{
		return const_iterator ( & first_, 0 );
	}

	const_iterator end() const
	{
		return const_iterator ( 0, & last_ );
	}

	single_list1()
		: last_(0)
	{}

	explicit single_list1 ( T value )
		: first_ ( value )
		, last_ ( 0 )
	{}

	~single_list1()
	{
		//node * next = first.next_;
		//while ( next )
		//{
		//	node * new_next = next->next_;
		//	delete next;
		//	next = new_next;
		//}
	}

	inline void clear()
	{
		node * next = first_.next_;
		while ( next )
		{
			node * new_next = next->next_;
			delete next;
			next = new_next;
		}
		first_.value_ = T();
		first_.next_ = 0;
		last_ = 0;
	}

	inline void setAutoCopy ( node & rhs )
	{
		first_.value_ = rhs.first_.value_;
		first_.next_ = rhs.first_.next_;
		last_ = rhs.last_;
		rhs.first_.value_ = T();
		rhs.first_.next_ = 0;
		rhs.last_ = 0;
	}

	inline bool empty() const
	{
		return first_.value_ == 0 && first_.next_ == 0;
	}

	inline void setFirst ( T value )
	{
		//XASSERT ( !first.next );
		//XASSERT ( !last );
		first_.value_ = value;
	}

	inline void push_back ( T value )
	{
		if ( !first_.value_ )
		{
			first_.value_ = value;
		}
		else/* if ( first.value_ )*/
		{
			node * next = new node ( value );
			if ( last_ )
				last_->next_ = next;
			else
				first_.next_ = next;
			last_ = next;
		}
	}

	inline void push_back2 ( T value )
	{
		//if ( !first_.value_ )
		//{
		//	first_.value_ = value;
		//}
		//else/* if ( first.value_ )*/
		{
			node * next = new node ( value );
			if ( last_ )
				last_->next_ = next;
			else
				first_.next_ = next;
			last_ = next;
		}
	}

	inline T getLast() const
	{
		if ( last_ )
		{
			return last_->value_;
		}
		else
		{
			return first_.value_;
		}
	}

	inline void delete_nodes()
	{
		node * next = first_.next_;
		while ( next )
		{
			node * old_next = next;
			next = next->next_;
			delete old_next;
		}
	}

	inline void delete_nodes_and_values()
	{
		//XASSERT ( this != (void*)0x00328558 );
		delete first_.value_;
		node * next = first_.next_;
		while ( next )
		{
			delete next->value_;
			node * old_next = next;
			next = next->next_;
			delete old_next;
		}
	}

	unsigned eval_size() const
	{
		unsigned size = 0;
		const node * next = & first_;
		while ( next )
		{
			if ( next->value_ )
				++size;
			next = next->next_;
		}
		return size;
	}

	void insert ( iterator it, T value )
	{
		if ( it.current_ == 0 ) // adding to the end of list
		{
			push_back ( value );
		}
		else
		{
			if ( it.prev_ == 0 ) // add to the begin of list
			{
				node * next = new node ( first_ );
				first_.value_ = value;
				first_.next_ = next;
				if ( !last_ )
					last_ = next;
			}
			else
			{
				node * next = new node ( value, it.current_ );
				it.prev_->next_ = next;
			}
		}
	}

	void erase ( iterator it )
	{
		//if ( it.current_ == 0 ) // erase from the end of list
		//{
		//	// empty
		//}
		//else
		{
			if ( it.prev_ == 0 ) // erase from the begin of list
			{
				if ( first_.next_ )
				{
					first_ = first_->next_;
					delete first_->next_;
					if ( it.current_->next_ )
						last_ = 0;
				}
				else
				{
					first_.value_ = T();
				}
			}
			else // erase medium element
			{
				it.prev_->next_ = it.current_->next;
				if ( it.current_ == last_ ) // erase last element
				{
					last_ = it.prev_;
				}
				delete it.current_;
			}
		}
	}

	iterator find ( T value )
	{
		if ( first_.value_ == value )
		{
			return iterator ( & first_.value_, 0 );
		}
		node * prev = & first_.next_;
		node * next = first_.next_;
		while ( next )
		{
			if ( next->value_ == value )
			{
				return iterator ( next, prev );
			}
			prev = next;
			next = next->next_;
		}
		return iterator ( 0, 0 );
	}

	node first_;
	node * last_;

//private:
//	single_list1 ( const single_list1 & rhs );
//	single_list1& operator= ( const single_list1 & rhs );
};

//------------------------------------------------------------------------------
// value_single_list1
//------------------------------------------------------------------------------

template < typename T >
struct value_single_list1
{
	struct node
	{
		node()
			: value_(T())
			, next_(0)
		{}

		node ( T value )
			: value_(value)
			, next_(0)
		{}


		inline void * operator new ( size_t size );
		inline void operator delete ( void * address );

		//inline void * operator new ( size_t size )
		//{
		//	return SingleList1TemplateAllocator::allocate();
		//}

		//inline void operator delete ( void * address )
		//{
		//	SingleList1TemplateAllocator::deallocate ( address );
		//}


		T value_;
		node * next_;
	};

	class iterator
	{
	public:
		//template<class _Ty>
		//struct iterator_traits//<_Ty *>
		//{	// get traits from pointer
		typedef std::forward_iterator_tag iterator_category;
		typedef T value_type;
		typedef ptrdiff_t difference_type;
		typedef ptrdiff_t distance_type;	// retained
		typedef T * pointer;
		typedef T & reference;
		//};

	public:
		iterator ( node * current, node ** prev )
			: current_ ( current )
			, prev_ ( prev )
		{}

		iterator()
			: current_ ( 0 )
			, prev_ ( 0 )
		{}

		T & operator*() const
		{
			return current_->value_;
		}

		T * operator->() const
		{
			return & current_->value_;
		}

		bool operator== ( const iterator & rhs ) const
		{
			return current_ == rhs.current_;
		}

		bool operator!= ( const iterator & rhs ) const
		{
			return current_ != rhs.current_;
		}

		iterator & operator++ () // preincrement
		{
			prev_ = & current_;
			current_ = current_->next_;
			return *this;
		}

		iterator operator++ ( int ) // postincrement
		{
			iterator retValue = *this;
			prev_ = & current_;
			current_ = current_->next_;
			return retValue;
		}

	private:
		node * current_;
		node ** prev_;
	};

	class const_iterator
	{
	public:
		//template<class _Ty>
		//struct iterator_traits//<_Ty *>
		//{	// get traits from pointer
		typedef std::forward_iterator_tag iterator_category;
		typedef const T value_type;
		typedef ptrdiff_t difference_type;
		typedef ptrdiff_t distance_type;	// retained
		typedef const T * pointer;
		typedef const T & reference;
		//};

	public:
		const_iterator ( const node * current, const node ** prev )
			: current_ ( current )
			, prev_ ( prev )
		{}

		const_iterator()
			: current_ ( 0 )
			, prev_ ( 0 )
		{}

		const T & operator*() const
		{
			return current_->value_;
		}

		const T * operator->() const
		{
			return & current_->value_;
		}

		bool operator== ( const const_iterator & rhs ) const
		{
			return current_ == rhs.current_;
		}

		bool operator!= ( const const_iterator & rhs ) const
		{
			return current_ != rhs.current_;
		}

		const_iterator & operator++ () // preincrement
		{
			prev_ = & current_;
			current_ = current_->next_;
			return *this;
		}

		const_iterator operator++ ( int ) // postincrement
		{
			const_iterator retValue = *this;
			prev_ = & current_;
			current_ = current_->next_;
			return retValue;
		}

	private:
		const node * current_;
		const node ** prev_;
	};

	iterator begin()
	{
		return iterator ( & first_, 0 );
	}

	iterator end()
	{
		return iterator ( 0, & last_ );
	}

	const_iterator begin() const
	{
		return const_iterator ( & first_, 0 );
	}

	const_iterator end() const
	{
		return const_iterator ( 0, & last_ );
	}

	value_single_list1()
		: last_(0)
	{}

	explicit value_single_list1 ( T value )
		: first_ ( value )
		, last_ ( 0 )
	{}

	~value_single_list1()
	{
		node * next = first_.next_;
		while ( next )
		{
			node * new_next = next->next_;
			delete next;
			next = new_next;
		}
	}

	inline void clear()
	{
		node * next = first_.next_;
		while ( next )
		{
			node * new_next = next->next_;
			delete next;
			next = new_next;
		}
		first_.value_ = T();
		first_.next_ = 0;
		last_ = 0;
	}

	inline void setAutoCopy ( value_single_list1<T> & rhs )
	{
		first_.value_ = rhs.first_.value_;
		first_.next_ = rhs.first_.next_;
		last_ = rhs.last_;
		rhs.first_.value_ = T();
		rhs.first_.next_ = 0;
		rhs.last_ = 0;
	}

	inline void setFirst ( T value )
	{
		XASSERT ( !first_.next_ );
		XASSERT ( !last_ );
		first_.value_ = value;
	}

	inline void push_back ( T value )
	{
		//if ( first.isNull() )
		//{
		//	first = value;
		//	return;
		//}
		node * next = new node ( value );
		//if ( !first.value_ )
		//{
		//	first.value_ = value;
		//}
		//else/* if ( first.value_ )*/
		{
			if ( last_ )
				last_->next_ = next;
			else
				first_.next_ = next;
			last_ = next;
		}
	}

	inline T & front()
	{
		return first_.value_;
	}

	inline const T & front() const
	{
		return first_.value_;
	}

	inline T & back()
	{
		if ( last_ )
		{
			return last_->value_;
		}
		else
		{
			return first_.value_;
		}
	}

	inline const T & back() const
	{
		if ( last_ )
		{
			return last_->value_;
		}
		else
		{
			return first_.value_;
		}
	}

	inline unsigned eval_size() const
	{
		unsigned size = 0;
		const node * next = & first_;
		while ( next )
		{
			++size;
			next = next->next_;
		}
		return size;
	}

	//inline void delete_nodes()
	//{
	//	//if ( first.value_ )
	//	{
	//		node * next = first.next_;
	//		while ( next )
	//		{
	//			node * old_next = next;
	//			next = next->next_;
	//			delete old_next;
	//		}
	//	}
	//}

	//inline void delete_nodes_and_values()
	//{
	//	//if ( first.value_ )
	//	{
	//		//first.value_.delete_value();
	//		node * next = first.next_;
	//		while ( next )
	//		{
	//			//next->value_.delete_value();
	//			node * old_next = next;
	//			next = next->next_;
	//			delete old_next;
	//		}
	//	}
	//}

	void insert ( iterator it, T value )
	{
		if ( it.current_ == 0 ) // adding to the end of list
		{
			push_back ( value );
		}
		else
		{
			if ( it.prev_ == 0 ) // add to the begin of list
			{
				node * next = new node ( first_ );
				first_.value_ = value;
				first_.next_ = next;
				if ( !last_ )
					last_ = next;
			}
			else
			{
				node * next = new node ( value, it.current_ );
				it.prev_->next_ = next;
			}
		}
	}

	void erase ( iterator it )
	{
		//if ( it.current_ == 0 ) // erase from the end of list
		//{
		//	// empty
		//}
		//else
		{
			if ( it.prev_ == 0 ) // erase from the begin of list
			{
				if ( first_.next_ )
				{
					first_ = first_->next_;
					delete first_->next_;
					if ( it.current_->next_ )
						last_ = 0;
				}
				else
				{
					first_.value_ = T();
				}
			}
			else // erase medium element
			{
				it.prev_->next_ = it.current_->next;
				if ( it.current_ == last_ ) // erase last element
				{
					last_ = it.prev_;
				}
				delete it.current_;
			}
		}
	}

	iterator find ( T value )
	{
		if ( first_.value_ == value )
		{
			return iterator ( & first_.value_, 0 );
		}
		node * prev = & first_.next_;
		node * next = first_.next_;
		while ( next )
		{
			if ( next->value_ == value )
			{
				return iterator ( next, prev );
			}
			prev = next;
			next = next->next_;
		}
		return iterator ( 0, 0 );
	}

	node first_;
	node * last_;

private:
	value_single_list1 ( const value_single_list1 & rhs );
	value_single_list1& operator= ( const value_single_list1 & rhs );
};

//------------------------------------------------------------------------------


template < typename T >
inline void * value_single_list1<T>::node::operator new ( size_t size )
{
	return FixedSmallObjAllocator
		< ALIGN_TO_2_POW_N_OR_64(sizeof(value_single_list1<T>::node)),
			DefaultAllocatorId,
			OwnPageAllocator >::allocate();
}

template < typename T >
inline void value_single_list1<T>::node::operator delete ( void * address )
{
	FixedSmallObjAllocator
		< ALIGN_TO_2_POW_N_OR_64(sizeof(value_single_list1<T>::node)),
			DefaultAllocatorId,
			OwnPageAllocator >::deallocate ( address );
}


//------------------------------------------------------------------------------
// value_single_list1_with_delete_value
//------------------------------------------------------------------------------

template < typename T >
struct value_single_list1_with_delete_value
{
	struct node
	{
		node()
			: value_(T())
			, next_(0)
		{}

		node ( T value )
			: value_(value)
			, next_(0)
		{}

		
		inline void * operator new ( size_t size );
		inline void operator delete ( void * address );

		//inline void * operator new ( size_t size )
		//{
		//	return SingleList1TemplateAllocator::allocate();
		//}

		//inline void operator delete ( void * address )
		//{
		//	SingleList1TemplateAllocator::deallocate ( address );
		//}
		

		T value_;
		node * next_;
	};

	class iterator
	{
	public:
		//template<class _Ty>
		//struct iterator_traits//<_Ty *>
		//{	// get traits from pointer
		typedef std::forward_iterator_tag iterator_category;
		typedef T value_type;
		typedef ptrdiff_t difference_type;
		typedef ptrdiff_t distance_type;	// retained
		typedef T * pointer;
		typedef T & reference;
		//};

	public:
		iterator ( node * current, node ** prev )
			: current_ ( current )
			, prev_ ( prev )
		{}

		iterator()
			: current_ ( 0 )
			, prev_ ( 0 )
		{}

		T & operator*() const
		{
			return current_->value_;
		}

		T * operator->() const
		{
			return & current_->value_;
		}

		bool operator== ( const iterator & rhs ) const
		{
			return current_ == rhs.current_;
		}

		bool operator!= ( const iterator & rhs ) const
		{
			return current_ != rhs.current_;
		}

		iterator & operator++ () // preincrement
		{
			prev_ = & current_;
			current_ = current_->next_;
			return *this;
		}

		iterator operator++ ( int ) // postincrement
		{
			iterator retValue = *this;
			prev_ = & current_;
			current_ = current_->next_;
			return retValue;
		}

	private:
		node * current_;
		node ** prev_;
	};

	class const_iterator
	{
	public:
		//template<class _Ty>
		//struct iterator_traits//<_Ty *>
		//{	// get traits from pointer
		typedef std::forward_iterator_tag iterator_category;
		typedef const T value_type;
		typedef ptrdiff_t difference_type;
		typedef ptrdiff_t distance_type;	// retained
		typedef const T * pointer;
		typedef const T & reference;
		//};

	public:
		const_iterator ( const node * current, const node ** prev )
			: current_ ( current )
			, prev_ ( prev )
		{}

		const_iterator()
			: current_ ( 0 )
			, prev_ ( 0 )
		{}

		const T & operator*() const
		{
			return current_->value_;
		}

		const T * operator->() const
		{
			return & current_->value_;
		}

		bool operator== ( const const_iterator & rhs ) const
		{
			return current_ == rhs.current_;
		}

		bool operator!= ( const const_iterator & rhs ) const
		{
			return current_ != rhs.current_;
		}

		const_iterator & operator++ () // preincrement
		{
			prev_ = & current_;
			current_ = current_->next_;
			return *this;
		}

		const_iterator operator++ ( int ) // postincrement
		{
			const_iterator retValue = *this;
			prev_ = & current_;
			current_ = current_->next_;
			return retValue;
		}

	private:
		const node * current_;
		const node ** prev_;
	};

	iterator begin()
	{
		return iterator ( & first_, 0 );
	}

	iterator end()
	{
		return iterator ( 0, & last_ );
	}

	const_iterator begin() const
	{
		return const_iterator ( & first_, 0 );
	}

	const_iterator end() const
	{
		return const_iterator ( 0, & last_ );
	}

	value_single_list1_with_delete_value()
		: last_(0)
	{}

	explicit value_single_list1_with_delete_value ( T value )
		: first_ ( value )
		, last_ ( 0 )
	{}

	~value_single_list1_with_delete_value()
	{
		//if ( first.value_ )
		{
			first_.value_.delete_value();
			node * next = first_.next_;
			while ( next )
			{
				next->value_.delete_value();
				node * old_next = next;
				next = next->next_;
				delete old_next;
			}
		}
	}

	inline void clear()
	{
		first_.value_.clear_value();
		node * next = first_.next_;
		while ( next )
		{
			next->value_.clear_value();
			node * old_next = next;
			next = next->next_;
			delete old_next;
		}
		first_.value_ = T();
		first_.next_ = 0;
		last_ = 0;
		//node * next = first_.next_;
		//while ( next )
		//{
		//	node * new_next = next->next_;
		//	delete next;
		//	next = new_next;
		//}
		//first_.value_ = T();
		//first_.next_ = 0;
		//last_ = 0;
	}

	inline void delete_nodes_and_values()
	{
		first_.value_.delete_value();
		node * next = first_.next_;
		while ( next )
		{
			next->value_.delete_value();
			node * old_next = next;
			next = next->next_;
			delete old_next;
		}
		first_.value_ = T();
		first_.next_ = 0;
		last_ = 0;
	}

	inline void setAutoCopy ( value_single_list1_with_delete_value<T> & rhs )
	{
		first_.value_ = rhs.first_.value_;
		first_.next_ = rhs.first_.next_;
		last_ = rhs.last_;
		rhs.first_.value_.clear_value();
		rhs.first_.next_ = 0;
		rhs.last_ = 0;
	}

	inline void setFirst ( T value )
	{
		//XASSERT ( !first.next );
		//XASSERT ( !last );
		first_.value_ = value;
	}

	inline void push_back ( T value )
	{
		node * next = new node ( value );
		//if ( !first.value_ )
		//{
		//	first.value_ = value;
		//}
		//else/* if ( first.value_ )*/
		{
			if ( last_ )
				last_->next_ = next;
			else
				first_.next_ = next;
			last_ = next;
		}
	}

	inline const T & back() const
	{
		if ( last_ )
		{
			return last_->value_;
		}
		else
		{
			return first_.value_;
		}
	}

	inline T & back()
	{
		if ( last_ )
		{
			return last_->value_;
		}
		else
		{
			return first_.value_;
		}
	}

	inline T getLast() const
	{
		if ( last_ )
		{
			return last_->value_;
		}
		else
		{
			return first_.value_;
		}
	}

	inline bool empty() const
	{
		return first_.value_.value_empty() && last_ == 0;
	}

	//inline void delete_nodes()
	//{
	//	if ( first.value_ )
	//	{
	//		node * next = first.next_;
	//		while ( next )
	//		{
	//			node * old_next = next;
	//			next = next->next_;
	//			delete old_next;
	//		}
	//	}
	//}

	//inline void delete_nodes_and_values()
	//{
	//	//if ( first.value_ )
	//	{
	//		first.value_.delete_value();
	//		node * next = first.next_;
	//		while ( next )
	//		{
	//			next->value_.delete_value();
	//			node * old_next = next;
	//			next = next->next_;
	//			delete old_next;
	//		}
	//	}
	//}

	void insert ( iterator it, T value )
	{
		if ( it.current_ == 0 ) // adding to the end of list
		{
			push_back ( value );
		}
		else
		{
			if ( it.prev_ == 0 ) // add to the begin of list
			{
				node * next = new node ( first_ );
				first_.value_ = value;
				first_.next_ = next;
				if ( !last_ )
					last_ = next;
			}
			else
			{
				node * next = new node ( value, it.current_ );
				it.prev_->next_ = next;
			}
		}
	}

	void erase ( iterator it )
	{
		//if ( it.current_ == 0 ) // erase from the end of list
		//{
		//	// empty
		//}
		//else
		{
			if ( it.prev_ == 0 ) // erase from the begin of list
			{
				if ( first_.next_ )
				{
					first_ = first_->next_;
					first_->next_->delete_value();
					delete first_->next_;
					if ( it.current_->next_ )
						last_ = 0;
				}
				else
				{
					first_.value_ = T();
				}
			}
			else // erase medium element
			{
				it.prev_->next_ = it.current_->next;
				if ( it.current_ == last_ ) // erase last element
				{
					last_ = it.prev_;
				}
				it.current_->delete_value();
				delete it.current_;
			}
		}
	}

	iterator find ( T value )
	{
		if ( first_.value_ == value )
		{
			return iterator ( & first_.value_, 0 );
		}
		node * prev = & first_.next_;
		node * next = first_.next_;
		while ( next )
		{
			if ( next->value_ == value )
			{
				return iterator ( next, prev );
			}
			prev = next;
			next = next->next_;
		}
		return iterator ( 0, 0 );
	}

	node first_;
	node * last_;

private:
	value_single_list1_with_delete_value ( const value_single_list1_with_delete_value & rhs );
	value_single_list1_with_delete_value& operator= ( const value_single_list1_with_delete_value & rhs );
};

//------------------------------------------------------------------------------


template < typename T >
inline void * value_single_list1_with_delete_value<T>::node::operator new ( size_t size )
{
	return FixedSmallObjAllocator
		< ALIGN_TO_2_POW_N_OR_64(sizeof(value_single_list1<T>::node)),
			DefaultAllocatorId,
			OwnPageAllocator >::allocate();
}

template < typename T >
inline void value_single_list1_with_delete_value<T>::node::operator delete ( void * address )
{
	FixedSmallObjAllocator
		< ALIGN_TO_2_POW_N_OR_64(sizeof(value_single_list1<T>::node)),
			DefaultAllocatorId,
			OwnPageAllocator >::deallocate ( address );
}


//----------------------------------------------------------
// IdentifierSingleList1Allocator
//----------------------------------------------------------


//struct identifier_single_list1_size
//{
//	byte bytes[8];
//};
//
//typedef FixedSmallObjAllocator
//	< identifier_single_list1_size, 4, OwnPageAllocator >
//		IdentifierSingleList1Allocator;

typedef FixedSmallObjAllocator
	< 8, DefaultAllocatorId, OwnPageAllocator >
		IdentifierSingleList1Allocator;


//------------------------------------------------------------------------------
// IdentifierSingleList
//------------------------------------------------------------------------------

struct IdentifierSingleList
{
	struct node
	{
		node()
			: next_(0)
		{}

		node ( WPHashString id )
			: identifier(id)
			, next_(0)
		{}


		inline void * operator new ( size_t size )
		{
			return IdentifierSingleList1Allocator::allocate();
		}

		inline void operator delete ( void * address )
		{
			IdentifierSingleList1Allocator::deallocate ( address );
		}


		PHashString identifier;
		node * next_;
	};

	IdentifierSingleList()
		: last_(0)
	{}

	explicit IdentifierSingleList ( WPHashString phs )
		: last_(0)
	{
		setFirst ( phs );
	}

	~IdentifierSingleList()
	{
		node * next = first_.next_;
		while ( next )
		{
			node * new_next = next->next_;
			delete next;
			next = new_next;
		}
	}

	inline void clear()
	{
		node * next = first_.next_;
		while ( next )
		{
			node * new_next = next->next_;
			delete next;
			next = new_next;
		}
		first_.identifier = WPHashString();
		first_.next_ = 0;
		last_ = 0;
	}

	inline void setAutoCopy ( IdentifierSingleList & rhs )
	{
		first_.identifier = rhs.first_.identifier;
		first_.next_ = rhs.first_.next_;
		last_ = rhs.last_;
		rhs.first_.identifier = WPHashString();
		rhs.first_.next_ = 0;
		rhs.last_ = 0;
	}

	inline void setFirst ( WPHashString id )
	{
		//removeAll(); // ?
		first_.identifier = id;
	}

	inline void push_back ( WPHashString id )
	{
		node * next = new node ( id );
		if ( last_ )
			last_->next_ = next;
		else
			first_.next_ = next;
		last_ = next;
	}

	inline bool isCompound() const
	{
		return first_.next_ != 0;
	}

	inline WPHashString getFirst() const
	{
		return first_.identifier;
	}

	inline WPHashString getLast() const
	{
		if ( last_ )
			return last_->identifier;
		else
			return first_.identifier;
	}

	//PHashString generateDottedName() const
	//{
	//	unsigned len = -1;
	//	const IdentifierSingleListStruct * next = & first;
	//	while ( next )
	//	{
	//		len += getLength ( next->identifier );
	//		++len;
	//		next = next->next;
	//	}
	//	if ( len > string_buffer_size )
	//	{
	//		if ( string_buffer )
	//			delete[] string_buffer;
	//		if ( string_buffer_size )
	//		{
	//			string_buffer_size *= 2;
	//		}
	//		else
	//		{
	//			string_buffer_size = string_buffer_start_size;
	//		}
	//		string_buffer = new char[string_buffer_size];
	//	}
	//	else if ( len == -1 )
	//	{
	//		return 0;
	//	}
	//	char * p_string_buffer = & string_buffer[0];
	//	next = & first;
	//	while ( next )
	//	{
	//		unsigned len_i = getLength ( next->identifier );
	//		memcpy ( p_string_buffer, getString ( next->identifier ), len_i );
	//		p_string_buffer += len_i;
	//		next = next->next;
	//		if ( next )
	//		{
	//			*p_string_buffer = '.';
	//			++p_string_buffer;
	//		}
	//	}
	//	return PHashString ( (const Char*) string_buffer, (const Char*) & string_buffer[len] );
	//}

	//PHashString generateDottedName ( WPHashString last_identifier ) const
	//{
	//	unsigned len = getLength ( last_identifier );
	//	const IdentifierSingleListStruct * next = & first;
	//	while ( next && next->identifier )
	//	{
	//		len += getLength ( next->identifier );
	//		++len;
	//		next = next->next;
	//	}
	//	if ( len > string_buffer_size )
	//	{
	//		if ( string_buffer )
	//			delete[] string_buffer;
	//		if ( string_buffer_size )
	//		{
	//			string_buffer_size *= 2;
	//		}
	//		else
	//		{
	//			string_buffer_size = string_buffer_start_size;
	//		}
	//		string_buffer = new char[string_buffer_size];
	//	}
	//	else if ( len == -1 )
	//	{
	//		return 0;
	//	}
	//	char * p_string_buffer = & string_buffer[0];
	//	next = & first;
	//	while ( next && next->identifier )
	//	{
	//		unsigned len_i = getLength ( next->identifier );
	//		memcpy ( p_string_buffer, getString ( next->identifier ), len_i );
	//		p_string_buffer += len_i;
	//		*p_string_buffer = '.';
	//		++p_string_buffer;
	//		next = next->next;
	//	}
	//	unsigned len_i = getLength ( last_identifier );
	//	memcpy ( p_string_buffer, getString ( last_identifier ), len_i );
	//	p_string_buffer += len_i;
	//	return PHashString ( (const Char*) string_buffer, (const Char*) & string_buffer[len] );
	//}



	node first_;
	node * last_;

	//static const unsigned string_buffer_start_size = 64;
	//static char * string_buffer;
	//static unsigned string_buffer_size;

private:
	IdentifierSingleList ( const IdentifierSingleList & rhs );
	IdentifierSingleList& operator= ( const IdentifierSingleList & rhs );
};

//------------------------------------------------------------------------------

#endif // _SINGLE_LIST_H_
