// RefactoringRename.cpp

#include "stdafx.h"

#include "RefactoringRename.h"
#include "CsEntity.h"
#include "CsParserTypes.h"
#include "ICsFile.h"
#include "FileManager.h"


//------------------------------------------------------------------------------
// RefactoringRename
//------------------------------------------------------------------------------

void RefactoringRename::setProjectData (
	CsEntityNamespace * global_namespace,
	const std::deque<ICsFile*> & files
)
{
	global_namespace_	= global_namespace;
	std::deque<ICsFile*>::const_iterator it = files.begin();
	std::deque<ICsFile*>::const_iterator it_end = files.end();
	for ( ; it != it_end; ++it )
	{
		files_.insert ( std::make_pair ( (*it)->getPath(), *it ) );
	}
}

//------------------------------------------------------------------------------

void RefactoringRename::initBeforeRefactoring()
{
	new_name_ = PHashString();
	method_node_ = 0;
	bNeedGlobalEntityReparsing_ = false;
	bAssemblyMethod_ = false;
	method_set_.clear();
	type_lookup_map_.clear();
	bAssemblyProperty_ = false;
	property_set_.clear();
}

//------------------------------------------------------------------------------

void RefactoringRename::rename ( const char * path, unsigned line, const char * old_identifier, unsigned identifier_number_in_line, const char * new_identifier )
{
	initBeforeRefactoring();
	PHashString phs_old = PHashString ( (const Char*)old_identifier );
	new_name_ = PHashString ( (const Char*)new_identifier );
	ICsFile * file = findFile ( path );
	if ( !file )
	{
		ASSERT(0);
		return;
	}
	CsCompilationUnit * unit = file->getCompilationUnit();
	if ( !unit )
	{
		ASSERT(0);
		return;
	}
	MediumFastDeque<CsToken>::iterator it = file->getTokens().begin();
	MediumFastDeque<CsToken>::iterator it_end = file->getTokens().end();
	bool bFound = false;
	unsigned cur_line = 1;
	unsigned cur_identifier_number_in_line = 0;
	for ( ; it != it_end && cur_line <= line; ++it )
	{
		if ( it->type == tkNEW_LINE )
		{
			++cur_line;
			//if ( cur_line == line )
			//	cur_identifier_number_in_line = 0;
		}
		if ( cur_line == line )
		{
			if ( it->type == tkIDENTIFIER
				&& HashString::equal ( it->getWPHashString(), phs_old ) )
			{
				if ( cur_identifier_number_in_line == identifier_number_in_line )
					break;
				else
					++cur_identifier_number_in_line;
			}
		}
	}

	if ( it == it_end )
		return;

	CsTokenIndex token_index = it->index;
	rename ( file, token_index );
}

//------------------------------------------------------------------------------

void RefactoringRename::rename ( const char * path, unsigned begin, unsigned end, const char * name )
{
	initBeforeRefactoring();
	new_name_ = PHashString ( (const Char*)name );
	ICsFile * file = findFile ( path );
	if ( !file )
	{
		ASSERT(0);
		return;
	}
	MediumFastDeque<CsToken>::iterator it = getToken ( file, begin, end );
	if ( it == file->getTokens().end() )
		return;
	if ( it->type != tkIDENTIFIER )
		return;
	CsTokenIndex token_index = it->index;

	rename ( file, token_index );
}

//------------------------------------------------------------------------------

void RefactoringRename::rename ( ICsFile * file, CsTokenIndex token_index )
{
	CsCompilationUnit * unit = file->getCompilationUnit();
	if ( !unit )
		return;
	CsFindTokenVisitor visitor;
	CsNode * node = visitor.findNodeByTokenIndex ( unit, token_index );
	if ( !node )
		return;

	switch ( node->type )
	{
		case n_local_variable_declarator:
			{
				renameLocalEntity ( static_cast<CsLocalVariableDeclarator*>(node)->entity );
			}
			break;
		case n_local_constant_declarator:
			{
				renameLocalEntity ( static_cast<CsLocalConstantDeclarator*>(node)->entity );
			}
			break;
		case n_simple_name:
			{
				CsSimpleName * simple_name = static_cast<CsSimpleName*>(node);
				if ( simple_name->entity )
				{
					switch ( simple_name->entity->e )
					{
						case ent_local_variable:
						case ent_local_constant:
						case ent_block_variable:
							{
								renameLocalEntity ( simple_name->entity );
							}
							break;
						case ent_formal_parameter:
							{
								renameLocalEntity ( simple_name->entity );
							}
							break;
						case ent_variable:
						case ent_constant:
							{
								renameVariableOrConstant ( simple_name->entity );
							}
							break;
						case ent_property:
							{
								renameProperty ( static_cast<CsEntityProperty*>(simple_name->entity) );
							}
							break;
						case ent_event:
							{
								renameEvent ( static_cast<CsEntityEvent*>(simple_name->entity) );
							}
							break;
						default:
							ASSERT(0);
					}
				}
				else
				{
					if ( simple_name->parent
						&& simple_name->parent->type == n_invocation_expression )
					{
						CsInvocationExpression * expr = static_cast<CsInvocationExpression*>(node);
						if ( expr->entity
							&& expr->entity->e == ent_method )
						{
							renameMethod ( static_cast<CsEntityMethod*>(expr->entity) );
						}
						else
							ASSERT(0);
					}
					else
						ASSERT(0);
				}
			}
			break;
		case n_foreach_statement:
			{
				CsForeachStatement * stat = static_cast<CsForeachStatement*>(node);
				renameLocalEntity ( stat->entity );
			}
			break;
		case n_catch_clause:
			{
				CsCatchClause * stat = static_cast<CsCatchClause*>(node);
				renameLocalEntity ( stat->entity );
			}
			break;
		case n_formal_parameter:
			{
				CsFormalParameter * param = static_cast<CsFormalParameter*>(node);
				renameLocalEntity ( param->entity );
			}
			break;
		case n_base_member_access:
			{
				CsBaseMemberAccess * expr = static_cast<CsBaseMemberAccess*>(node);
				if ( expr->entity )
				{
					switch ( expr->entity->e )
					{
						case ent_variable:
						case ent_constant:
							renameVariableOrConstant ( expr->entity );
							break;
						case ent_property:
							renameProperty ( static_cast<CsEntityProperty*>(expr->entity) );
							break;
						case ent_event:
							renameEvent ( static_cast<CsEntityEvent*>(expr->entity) );
							break;
						default:
							ASSERT(0);
					}
				}
			}
			break;
		case n_member_access:
			{
				CsMemberAccess * expr = static_cast<CsMemberAccess*>(node);
				if ( expr->entity )
				{
					switch ( expr->entity->e )
					{
						case ent_variable:
						case ent_constant:
							renameVariableOrConstant ( expr->entity );
							break;
						case ent_property:
							renameProperty ( static_cast<CsEntityProperty*>(expr->entity) );
							break;
						case ent_event:
							renameEvent ( static_cast<CsEntityEvent*>(expr->entity) );
							break;
						default:
							ASSERT(0);
					}
				}
			}
			break;
		case n_pointer_member_access:
			{
				CsPointerMemberAccess * expr = static_cast<CsPointerMemberAccess*>(node);
				if ( expr->entity )
				{
					switch ( expr->entity->e )
					{
						case ent_variable:
						case ent_constant:
							renameVariableOrConstant ( expr->entity );
							break;
						case ent_property:
							renameProperty ( static_cast<CsEntityProperty*>(expr->entity) );
							break;
						case ent_event:
							renameEvent ( static_cast<CsEntityEvent*>(expr->entity) );
							break;
						default:
							ASSERT(0);
					}
				}
			}
			break;
		case n_embedded_type_member_access:
			{
				CsEmbeddedTypeMemberAccess * expr = static_cast<CsEmbeddedTypeMemberAccess*>(node);
				if ( expr->entity )
				{
					switch ( expr->entity->e )
					{
						case ent_variable:
						case ent_constant:
							renameVariableOrConstant ( expr->entity );
							break;
						case ent_property:
							renameProperty ( static_cast<CsEntityProperty*>(expr->entity) );
							break;
						case ent_event:
							renameEvent ( static_cast<CsEntityEvent*>(expr->entity) );
							break;
						default:
							ASSERT(0);
					}
				}
			}
			break;
		case n_constant_declarator:
			{
				CsConstantDeclarator * decl = static_cast<CsConstantDeclarator*>(node);
				if ( decl->entity )
				{
					renameVariableOrConstant ( decl->entity );
				}
			}
			break;
		case n_variable_declarator:
			{
				CsVariableDeclarator * decl = static_cast<CsVariableDeclarator*>(node);
				if ( decl->entity )
				{
					renameVariableOrConstant ( decl->entity );
				}
			}
			break;
		case n_method_declaration:
			{
				CsMethod * decl = static_cast<CsMethod*>(node);
				if ( decl->entity )
				{
					renameMethod ( decl->entity );
				}
			}
			break;
		case n_property_declaration:
			{
				CsProperty * decl = static_cast<CsProperty*>(node);
				if ( decl->entity )
				{
					renameProperty ( decl->entity );
				}
			}
			break;
		case n_event_accessor:
			{
				CsEventDeclarator * decl = static_cast<CsEventDeclarator*>(node);
				if ( decl->entity )
				{
					renameEvent ( decl->entity );
				}
			}
			break;
		case n_class:
			{
				CsClass * decl = static_cast<CsClass*>(node);
				if ( decl->entity )
				{
					renameType ( static_cast<CsEntityType*>(decl->entity) );
				}
			}
			break;
		case n_struct:
			{
				CsStruct * decl = static_cast<CsStruct*>(node);
				if ( decl->entity )
				{
					renameType ( static_cast<CsEntityType*>(decl->entity) );
				}
			}
			break;
		case n_interface:
			{
				CsInterface * decl = static_cast<CsInterface*>(node);
				if ( decl->entity )
				{
					renameType ( static_cast<CsEntityType*>(decl->entity) );
				}
			}
			break;
		case n_enum:
			{
				CsEnum * decl = static_cast<CsEnum*>(node);
				if ( decl->entity )
				{
					renameType ( static_cast<CsEntityType*>(decl->entity) );
				}
			}
			break;
		case n_delegate:
			{
				CsDelegate * decl = static_cast<CsDelegate*>(node);
				if ( decl->entity )
				{
					renameType ( static_cast<CsEntityType*>(decl->entity) );
				}
			}
			break;
		case n_type_ref:
			{
				CsTypeRef * typeref = static_cast<CsTypeRef*>(node);
				if ( typeref->using_directive_used_in_type_lookup
					&& typeref->compound_identifier.first_.value_.identifier_token == token_index )
				{
					renameUsingDirectiveAlias ( typeref->using_directive_used_in_type_lookup );
					break;
				}
				CsEntity * entity = 0;
				CsEntityTypeRef * entity_typeref = & typeref->entity_typeref;
loop:
				switch ( entity_typeref->type )
				{
					case et_class:
					case et_valuetype:
						{
							entity = entity_typeref->u.entity;
							CsCompoundIdentifier::node * next
								= & typeref->compound_identifier.first_;
							unsigned n_identifier = 1;
							while ( next )
							{
								if ( next->value_.identifier_token == token_index )
									break;
								next = next->next_;
								++n_identifier;
							}
							unsigned n_identifiers = typeref->compound_identifier.eval_size() - n_identifier;
							while ( n_identifiers > 0 )
							{
								entity = entity->parent;
								--n_identifiers;
							}
						}
						break;
					case et_array:
						{
							entity_typeref = & entity_typeref->u.array_specifier->type;
							goto loop;
						}
						break;
				}
				if ( entity )
				{
					switch ( entity->e )
					{
						case ent_class:
						case ent_struct:
						case ent_interface:
						case ent_enum:
						case ent_delegate:
							renameType ( static_cast<CsEntityType*>(entity) );
							break;
						case ent_namespace:
							renameNamespace ( static_cast<CsEntityNamespace*>(entity) );
							break;
						default:
							ASSERT(0);
					}
				}
				else
					ASSERT(0);
			}
			break;
	case n_using_alias_directive:
		{
			CsUsingAliasDirective * ud = static_cast<CsUsingAliasDirective*>(node);
			if ( ud->identifier_token == token_index )
			{
				renameUsingDirectiveAlias ( ud );
			}
			else if ( ud->using_directive_used_in_type_lookup
				&& ud->namespace_or_type_name.first_.value_.identifier_token == token_index )
			{
				renameUsingDirectiveAlias ( ud->using_directive_used_in_type_lookup );
			}
			else
			{
				CsEntity * entity = ud->namespace_or_type_entity;
				CsCompoundIdentifier::node * next
					= & ud->namespace_or_type_name.first_;
				unsigned n_identifier = 1;
				while ( next )
				{
					if ( next->value_.identifier_token == token_index )
						break;
					next = next->next_;
					++n_identifier;
				}
				unsigned n_identifiers = ud->namespace_or_type_name.eval_size() - n_identifier;
				while ( n_identifiers > 0 )
				{
					entity = entity->parent;
					--n_identifiers;
				}
				if ( entity )
				{
					switch ( entity->e )
					{
						case ent_class:
						case ent_struct:
						case ent_interface:
						case ent_enum:
						case ent_delegate:
							renameType ( static_cast<CsEntityType*>(entity) );
							break;
						case ent_namespace:
							renameNamespace ( static_cast<CsEntityNamespace*>(entity) );
							break;
						default:
							ASSERT(0);
					}
				}
				else
					ASSERT(0);
			}
		}
		break;
	case n_using_namespace_directive:
		{
			CsUsingNamespaceDirective * ud = static_cast<CsUsingNamespaceDirective*>(node);
			{
				CsEntity * entity = ud->namespace_or_type_entity;
				CsCompoundIdentifier::node * next
					= & ud->namespace_or_type_name.first_;
				unsigned n_identifier = 1;
				while ( next )
				{
					if ( next->value_.identifier_token == token_index )
						break;
					next = next->next_;
					++n_identifier;
				}
				unsigned n_identifiers = ud->namespace_or_type_name.eval_size() - n_identifier;
				while ( n_identifiers > 0 )
				{
					entity = entity->parent;
					--n_identifiers;
				}
				if ( entity )
				{
					switch ( entity->e )
					{
						case ent_namespace:
							renameNamespace ( static_cast<CsEntityNamespace*>(entity) );
							break;
						default:
							ASSERT(0);
					}
				}
				else
					ASSERT(0);
			}
		}
		break;
		//case n_using_directive:
		//	{
		//		CsUsingDirective * using_directive = static_cast<CsUsingDirective*>(node);
		//		if ( using_directive->bUsingAliasDirective
		//			&& using_directive->identifier_token == token_index )
		//		{
		//			renameUsingDirectiveAlias ( using_directive );
		//		}
		//		else if ( using_directive->using_directive_used_in_type_lookup
		//			&& using_directive->namespace_or_type_name.first_.value_.identifier_token == token_index )
		//		{
		//			renameUsingDirectiveAlias ( using_directive->using_directive_used_in_type_lookup );
		//		}
		//		else
		//		{
		//			CsEntity * entity = using_directive->namespace_or_type_entity;
		//			CsCompoundIdentifier::node * next
		//				= & using_directive->namespace_or_type_name.first_;
		//			unsigned n_identifier = 1;
		//			while ( next )
		//			{
		//				if ( next->value_.identifier_token == token_index )
		//					break;
		//				next = next->next_;
		//				++n_identifier;
		//			}
		//			unsigned n_identifiers = using_directive->namespace_or_type_name.eval_size() - n_identifier;
		//			while ( n_identifiers > 0 )
		//			{
		//				entity = entity->parent;
		//				--n_identifiers;
		//			}
		//			if ( entity )
		//			{
		//				switch ( entity->e )
		//				{
		//					case ent_class:
		//					case ent_struct:
		//					case ent_interface:
		//					case ent_enum:
		//					case ent_delegate:
		//						renameType ( static_cast<CsEntityType*>(entity) );
		//						break;
		//					case ent_namespace:
		//						renameNamespace ( static_cast<CsEntityNamespace*>(entity) );
		//						break;
		//					default:
		//						ASSERT(0);
		//				}
		//			}
		//			else
		//				ASSERT(0);
		//		}
		//	}
		//	break;
		case n_namespace:
			{
				CsNamespace * decl = static_cast<CsNamespace*>(node);
				{
					CsEntity * entity = decl->entity;
					CsCompoundIdentifier::node * next
						= & decl->compound_identifier.first_;
					unsigned n_identifier = 1;
					while ( next )
					{
						if ( next->value_.identifier_token == token_index )
							break;
						next = next->next_;
						++n_identifier;
					}
					unsigned n_identifiers = decl->compound_identifier.eval_size() - n_identifier;
					while ( n_identifiers > 0 )
					{
						entity = entity->parent;
						--n_identifiers;
					}
					if ( entity )
					{
						renameNamespace ( static_cast<CsEntityNamespace*>(entity) );
					}
					else
						ASSERT(0);
				}
			}
			break;
		case n_label:
			{
				CsLabel * label
					= static_cast<CsLabel*>(node);
				renameLabel ( label );
			}
			break;
		case n_goto_statement:
			{
				CsGotoStatement * stat
					= static_cast<CsGotoStatement*>(node);
				if ( stat->label_type == tkIDENTIFIER )
				{
					ASSERT ( stat->label_token == token_index );
					if ( stat->goto_target ) // otherwise compile error
					{
						ASSERT ( stat->goto_target->type == n_label );
						renameLabel ( static_cast<CsLabel*>(stat->goto_target) );
					}
				}
				else
					ASSERT(0);
			}
			break;
		default:
			ASSERT(0);
	}
}

//------------------------------------------------------------------------------

void RefactoringRename::renameLocalEntity ( CsEntity * entity )
{
	method_node_ = 0;
	bool bWarning = false;
	switch ( entity->e )
	{
		case ent_local_variable:
		case ent_local_constant:
			{
				CsScopeEntity * local_scope = static_cast<CsScopeEntity*>(entity->parent);
				if ( local_scope
					&& local_scope->scope_table.exist_id ( new_name_ ) )
				{
					bWarning = true; // local entity with new_name_ already defined in local scope
				}
			}
			break;
		case ent_block_variable:
			break;
		case ent_formal_parameter:
			{
				CsEntityFormalParameterList * param_list
					= static_cast<CsEntityFormalParameterList*>(entity->parent);
				single_list1<CsEntityFormalParameter*>::iterator it
					= param_list->entity_parameter_list->begin();
				single_list1<CsEntityFormalParameter*>::iterator it_end
					= param_list->entity_parameter_list->end();
				for ( ; it != it_end; ++it )
				{
					if ( (*it) != entity
						&& HashString::equal ( (*it)->name, new_name_ ) )
					{
						bWarning = true; // formal parameter with new_name_ already defined in formal parameter list
					}
				}
			}
			break;
		default:
			ASSERT(0);
	}
	CsNode * lookup_node = 0;
	switch ( entity->e )
	{
		case ent_local_variable:
			{
				CsNode * node = static_cast<CsEntityLocalVariable*>(entity)->decl;
				if ( node )
				{
					lookup_node = node->parent->parent->parent;
					if ( node->type == n_local_variable_declarator )
					{
						CsLocalVariableDeclarator * decl
							= static_cast<CsLocalVariableDeclarator*>(node);
						if ( decl )
							addTokenToReplace ( getFile ( node ), decl->identifier_token );
						else
							ASSERT(0);
					}
					else if ( node->type == n_fixed_point_declarator )
					{
						CsFixedPointDeclarator * decl
							= static_cast<CsFixedPointDeclarator*>(node);
						if ( decl )
							addTokenToReplace ( getFile ( node ), decl->identifier_token );
						else
							ASSERT(0);
					}
					else
						ASSERT(0);
				}
			}
			break;
		case ent_local_constant:
			{
				CsLocalConstantDeclarator * decl = static_cast<CsEntityLocalConstant*>(entity)->decl;
				if ( decl )
				{
					lookup_node = decl->parent->parent->parent;
					addTokenToReplace ( getFile ( decl ), decl->identifier_token );
				}
				else
					ASSERT(0);
			}
			break;
		case ent_block_variable:
			{
				CsNode * node = static_cast<CsEntityBlockVariable*>(entity)->decl;
				if ( node->type == n_foreach_statement )
				{
					CsForeachStatement * stat
						= static_cast<CsForeachStatement*>(node);
					lookup_node = stat;
					addTokenToReplace ( getFile ( stat ), stat->identifier_token );
				}
				else if ( node->type == n_catch_clause )
				{
					CsCatchClause * stat
						= static_cast<CsCatchClause*>(node);
					lookup_node = stat;
					addTokenToReplace ( getFile ( stat ), stat->identifier_token );
				}
				else
					ASSERT(0);
			}
			break;
		case ent_formal_parameter:
			{
				CsFormalParameter * param = static_cast<CsEntityFormalParameter*>(entity)->param;
				lookup_node = param->parent;
				if ( lookup_node->type == n_formal_parameter_list )
					lookup_node = lookup_node->parent;
				addTokenToReplace ( getFile ( param ), param->identifier_token );
			}
			break;
		default:
			ASSERT(0);
	}
	if ( !lookup_node )
		return;
	CsNodeCollectorVisitor visitor;
	visitor.collect_node_flags_[n_simple_name] = true;
	visitor.collectNodes ( lookup_node );
	MediumFastDeque<CsNode*>::iterator it = visitor.nodes_.begin();
	MediumFastDeque<CsNode*>::iterator it_end = visitor.nodes_.end();
	for ( ; it != it_end; ++it )
	{
		CsSimpleName * expr = static_cast<CsSimpleName*>(*it);
		if ( equal ( expr->identifier, new_name_ ) )
		{
			if ( expr->entity )
			{
				switch ( expr->entity->e )
				{
					case ent_local_variable:
						{
							//if ( (*it)->entity )
							//{
							//	
							//}
						}
						break;
					case ent_local_constant:
						{
							//if ( )
						}
						break;
					default:
						bWarning = true; // warning: refactoring shall produce compilation error or change program logic
						break;
				}
			}
		}
		if ( expr->entity == entity )
		{
			addTokenToReplace ( getFile ( expr ), expr->identifier_token );
		}
	}
	//if ( bWarning )
	//	; // report warning
	method_node_ = getMethod ( lookup_node );
	replaceTokens ( true );
}

//------------------------------------------------------------------------------

void RefactoringRename::renameVariableOrConstant ( CsEntity * entity )
{
	CsNodeCollectorVisitor visitor;
	visitor.collect_node_flags_[n_simple_name] = true;
	visitor.collect_node_flags_[n_base_member_access] = true;
	visitor.collect_node_flags_[n_member_access] = true;
	visitor.collect_node_flags_[n_pointer_member_access] = true;
	visitor.collect_node_flags_[n_embedded_type_member_access] = true;
	visitor.collect_node_flags_[n_constant_declarator] = true;
	visitor.collect_node_flags_[n_variable_declarator] = true;
	visitor.bCheckEntities_ = true;
	visitor.entities_.insert ( entity );

	std::map<std::string,ICsFile*>::iterator fit = files_.begin();
	std::map<std::string,ICsFile*>::iterator fit_end = files_.end();
	for ( ; fit != fit_end; ++fit )
	{
		visitor.collectNodes ( fit->second->getCompilationUnit() );
	}
	MediumFastDeque<CsNode*>::iterator it = visitor.nodes_.begin();
	MediumFastDeque<CsNode*>::iterator it_end = visitor.nodes_.end();
	for ( ; it != it_end; ++it )
	{
		switch ( (*it)->type )
		{
			case n_simple_name:
				{
					CsSimpleName * node = static_cast<CsSimpleName*>(*it);
					addTokenToReplace ( getFile ( node ), node->identifier_token );
				}
				break;
			case n_base_member_access:
				{
					CsBaseMemberAccess * node = static_cast<CsBaseMemberAccess*>(*it);
					addTokenToReplace ( getFile ( node ), node->member_token );
				}
				break;
			case n_member_access:
				{
					CsMemberAccess * node = static_cast<CsMemberAccess*>(*it);
					addTokenToReplace ( getFile ( node ), node->member_token );
				}
				break;
			case n_pointer_member_access:
				{
					CsPointerMemberAccess * node = static_cast<CsPointerMemberAccess*>(*it);
					addTokenToReplace ( getFile ( node ), node->member_token );
				}
				break;
			case n_embedded_type_member_access:
				{
					CsEmbeddedTypeMemberAccess * node = static_cast<CsEmbeddedTypeMemberAccess*>(*it);
					addTokenToReplace ( getFile ( node ), node->member_token );
				}
				break;
			case n_constant_declarator:
				{
					CsConstantDeclarator * node = static_cast<CsConstantDeclarator*>(*it);
					addTokenToReplace ( getFile ( node ), node->identifier_token );
				}
				break;
			case n_variable_declarator:
				{
					CsVariableDeclarator * node = static_cast<CsVariableDeclarator*>(*it);
					addTokenToReplace ( getFile ( node ), node->identifier_token );
				}
				break;
			default:
				ASSERT(0);
		}
	}

	bNeedGlobalEntityReparsing_ = true;
	replaceTokens ( true );
}

//------------------------------------------------------------------------------

void RefactoringRename::renameMethod ( CsEntityMethod * method )
{
	if ( method->flags1 & f_generated )
		return;
	if ( method->assembly )
		return;
	PHashString ctor ( (const Char*) ".ctor" );
	if ( HashString::equal ( ctor, method->name ) )
		return;
	PHashString dtor ( (const Char*) ".dtor" );
	if ( HashString::equal ( dtor, method->name ) )
		return;
	if ( method->owner )
	{
		ASSERT(0); // internal error?
		return;
	}

	if ( !method->owner ) // properties and event already supported
		method_set_.insert ( method );
	collectTypeMethods
		( static_cast<CsEntityType*>(method->parent), 0, method, true );
	collectTypeMethods
		( static_cast<CsEntityType*>(method->parent), 0, method, false );

	if ( bAssemblyMethod_ )
		return;

	CsNodeCollectorVisitor visitor;
	visitor.collect_node_flags_[n_invocation_expression] = true;
	visitor.bCheckEntities_ = true;
	visitor.entities_ = method_set_;

	std::map<std::string,ICsFile*>::iterator fit = files_.begin();
	std::map<std::string,ICsFile*>::iterator fit_end = files_.end();
	for ( ; fit != fit_end; ++fit )
	{
		visitor.collectNodes ( fit->second->getCompilationUnit() );
	}
	MediumFastDeque<CsNode*>::iterator it = visitor.nodes_.begin();
	MediumFastDeque<CsNode*>::iterator it_end = visitor.nodes_.end();
	for ( ; it != it_end; ++it )
	{
		switch ( (*it)->type )
		{
			case n_invocation_expression:
				{
					CsInvocationExpression * expr = static_cast<CsInvocationExpression*>(*it);
					ASSERT ( method_set_.find ( static_cast<CsEntityMethod*>(expr->entity) ) != method_set_.end() );
					CsNode * node = expr->parent;
					switch ( node->type )
					{
						case n_simple_name:
							{
								CsSimpleName * node = static_cast<CsSimpleName*>(*it);
								addTokenToReplace ( getFile ( node ), node->identifier_token );
							}
							break;
						case n_base_member_access:
							{
								CsBaseMemberAccess * node = static_cast<CsBaseMemberAccess*>(*it);
								addTokenToReplace ( getFile ( node ), node->member_token );
							}
							break;
						case n_member_access:
							{
								CsMemberAccess * node = static_cast<CsMemberAccess*>(*it);
								addTokenToReplace ( getFile ( node ), node->member_token );
							}
							break;
						case n_pointer_member_access:
							{
								CsPointerMemberAccess * node = static_cast<CsPointerMemberAccess*>(*it);
								addTokenToReplace ( getFile ( node ), node->member_token );
							}
							break;
						case n_embedded_type_member_access:
							{
								CsEmbeddedTypeMemberAccess * node = static_cast<CsEmbeddedTypeMemberAccess*>(*it);
								addTokenToReplace ( getFile ( node ), node->member_token );
							}
							break;
						default:
							ASSERT(0);
					}
				}
				break;
			default:
				ASSERT(0);
		}
	}

	{
		std::set<CsEntity*>::iterator it = method_set_.begin(); 
		std::set<CsEntity*>::iterator it_end = method_set_.end(); 
		for ( ; it != it_end; ++it )
		{
			ASSERT ( (*it)->e == ent_method );
			CsEntityMethod * entity = static_cast<CsEntityMethod*>(*it);
			CsNode * node = entity->decl;
			switch ( node->type )
			{
				case n_method_declaration:
					{
						CsMethod * method = static_cast<CsMethod*>(node);
						addTokenToReplace ( getFile ( method ), method->compound_identifier.back().identifier_token );
					}
					break;
				default:
					ASSERT(0);
			}
		}
	}

	bNeedGlobalEntityReparsing_ = true;
	replaceTokens ( true );
}

//------------------------------------------------------------------------------
// bFromDerived - true: lookup from derived to base, false: lookup from base to derived

void RefactoringRename::collectTypeMethods	(
	CsEntityType * type,
	CsEntityType * from_type,
	CsEntityMethod * from_method,
	bool bFromDerived
)
{
	//type->traceScopeAndName();
	//from_method->traceScopeAndName();
	CsEntityType * new_from_type = type;
	CsEntityMethod * new_from_method = from_method;
	if ( bFromDerived )
	{
		if ( from_type )
			type_lookup_map_[type].setWasNotifiedFromDerived ( true );
	}
	else
	{
		if ( from_type )
		{
			if ( from_type->e == ent_interface )
				type_lookup_map_[type].setWasNotifiedFromBaseInterface ( true );
			else
				type_lookup_map_[type].setWasNotifiedFromBaseClass ( true );
		}
	}
	bool bStopBaseLookupInClass = false;
	bool bStopBaseLookupInInterfaces = false;
	bool bStopDerivedLookup = false;
	bool bBaseLookupInClass = false;
	bool bBaseLookupInInterfaces = false;
	bool bDerivedLookup = false;
	{
		const ScopeHashTable::hash_id_struct2 * p_his
			= type->scope_table.get_id_list ( from_method->name );
		while ( p_his )
		{
			if ( p_his->data_->e == ent_method
				&& static_cast<CsEntityMethod*>(p_his->data_)->specifier.equal ( from_method->specifier ) )
			{
				CsEntityMethod * method = static_cast<CsEntityMethod*>(p_his->data_);
				//type_lookup_map_[type].method = method;
				if ( method->assembly )
					bAssemblyMethod_ = true;
				if ( bFromDerived )
				{
					if ( method->isOverrideMethod() )
					{
						if ( !method->owner ) // properties and event already supported
							method_set_.insert ( method );
						bDerivedLookup = true;
					}
					else if ( method->isVirtualMethod() )
					{
						if ( !method->owner ) // properties and event already supported
							method_set_.insert ( method );
						if ( type->e != ent_interface )
							bStopBaseLookupInClass = true;
						bDerivedLookup = true;
						new_from_method = method;
						new_from_type = type;
					}
					else
					{
						bStopBaseLookupInClass = true;
						bStopBaseLookupInInterfaces = true;
					}
				}
				else
				{
					if ( method->isOverrideMethod() )
					{
						if ( !method->owner ) // properties and event already supported
							method_set_.insert ( method );
						bBaseLookupInClass = true;
						bBaseLookupInInterfaces = true;
					}
					else if ( method->isVirtualMethod() )
					{
						if ( !from_type
							|| from_type->e == ent_interface )
						{
							if ( !method->owner ) // properties and event already supported
								method_set_.insert ( method );
						}
						if ( type->e != ent_interface
							&& from_method->parent->e != ent_interface )
						{
							bStopDerivedLookup = true;
						}
						bBaseLookupInClass = true;
						bBaseLookupInInterfaces = true;
						new_from_method = method;
						new_from_type = type;
					}
					else
					{
						bStopDerivedLookup = true;
					}
				}
			}
			p_his = p_his->next_;
		}
	}

// interface mapping
	if ( from_type
		&& from_type->e != ent_interface )
	{
		bBaseLookupInInterfaces = true;
	}
	if ( from_type
		&& from_type->e == ent_interface )
	{
		bBaseLookupInClass = true;
		bBaseLookupInInterfaces = true;
	}

	if ( !type_lookup_map_[type].isBaseWasLookup() )
	{
		if ( bFromDerived && !bStopBaseLookupInClass
			|| bBaseLookupInClass )
		{
			type_lookup_map_[type].setBaseWasLookup ( true );
			CsEntityClass * base = type->base;
			if ( base )
			{
				if ( !type_lookup_map_[base].isWasNotifiedFromDerived() )
				{
					collectTypeMethods
						( base, new_from_type, new_from_method, true );
				}
			}
		}
		if ( bFromDerived && !bStopBaseLookupInInterfaces
			|| bBaseLookupInInterfaces )
		{
			type_lookup_map_[type].setBaseWasLookup ( true );
			switch ( type->e )
			{
				case ent_class:
				case ent_struct:
				case ent_interface:
					{
						CsEntityTypeWithInterfaces * type_wi
							= static_cast<CsEntityTypeWithInterfaces*>(type);
						if ( type_wi->interfaces )
						{
							single_list1<CsEntityInterface*>::node * next
								= & type_wi->interfaces->first_;
							while ( next )
							{
								CsEntityInterface * base = next->value_;
								if ( base )
								{
									if ( !type_lookup_map_[base].isWasNotifiedFromDerived() )
									{
										collectTypeMethods
											( base, new_from_type, new_from_method, true );
									}
								}
								next = next->next_;
							}
						}
					}
					break;
			}
		}
	}

	if ( !type_lookup_map_[type].isDerivedWasLookup() )
	{
    if ( !bFromDerived && !bStopDerivedLookup
			|| bDerivedLookup )
		{
			type_lookup_map_[type].setDerivedWasLookup ( true );
			if ( type->derived_types )
			{
				single_list1<CsEntityType*>::node * next
					= & type->derived_types->first_;
				while ( next )
				{
					CsEntityType * derived = next->value_;
					if ( type->e == ent_interface )
					{
						if ( !type_lookup_map_[derived].isWasNotifiedFromBaseInterface() )
						{
							collectTypeMethods
								( derived, new_from_type, new_from_method, false );
						}
					}
					else
					{
						if ( !type_lookup_map_[derived].isWasNotifiedFromBaseClass() )
						{
							collectTypeMethods
								( derived, new_from_type, new_from_method, false );
						}
					}
					next = next->next_;
				}
			}
			//if ( type->derived_interfaces )
			//{
			//	single_list1<CsEntityInterface*>::node * next
			//		= & type->derived_interfaces->first;
			//	while ( next )
			//	{
			//		CsEntityInterface * derived = next->value_;
			//		if ( type->e == ent_interface )
			//		{
			//			if ( !type_lookup_map_[derived].isWasNotifiedFromBaseInterface() )
			//			{
			//				collectTypeMethods
			//					( derived, new_from_type, new_from_method, false );
			//			}
			//		}
			//		else
			//		{
			//			if ( !type_lookup_map_[derived].isWasNotifiedFromBaseClass() )
			//			{
			//				collectTypeMethods
			//					( derived, new_from_type, new_from_method, false );
			//			}
			//		}
			//		next = next->next_;
			//	}
			//}
		}
	}
}

//------------------------------------------------------------------------------

void RefactoringRename::renameProperty ( CsEntityProperty * property )
{
	//if ( property->flags1 & f_generated )
	//	return;
	if ( property->assembly )
		return;

	collectTypeProperties
		( static_cast<CsEntityType*>(property->parent), 0, property, true );
	collectTypeProperties
		( static_cast<CsEntityType*>(property->parent), 0, property, false );

	if ( bAssemblyProperty_ )
		return;

	CsNodeCollectorVisitor visitor;
	visitor.collect_node_flags_[n_simple_name] = true;
	visitor.collect_node_flags_[n_base_member_access] = true;
	visitor.collect_node_flags_[n_member_access] = true;
	visitor.collect_node_flags_[n_pointer_member_access] = true;
	visitor.collect_node_flags_[n_embedded_type_member_access] = true;
	//visitor.collect_node_flags_[n_property_declaration] = true;
	visitor.bCheckEntities_ = true;
	visitor.entities_ = method_set_;

	std::map<std::string,ICsFile*>::iterator fit = files_.begin();
	std::map<std::string,ICsFile*>::iterator fit_end = files_.end();
	for ( ; fit != fit_end; ++fit )
	{
		visitor.collectNodes ( fit->second->getCompilationUnit() );
	}
	MediumFastDeque<CsNode*>::iterator it = visitor.nodes_.begin();
	MediumFastDeque<CsNode*>::iterator it_end = visitor.nodes_.end();
	for ( ; it != it_end; ++it )
	{
		switch ( (*it)->type )
		{
			case n_simple_name:
				{
					CsSimpleName * node = static_cast<CsSimpleName*>(*it);
					addTokenToReplace ( getFile ( node ), node->identifier_token );
				}
				break;
			case n_base_member_access:
				{
					CsBaseMemberAccess * node = static_cast<CsBaseMemberAccess*>(*it);
					addTokenToReplace ( getFile ( node ), node->member_token );
				}
				break;
			case n_member_access:
				{
					CsMemberAccess * node = static_cast<CsMemberAccess*>(*it);
					addTokenToReplace ( getFile ( node ), node->member_token );
				}
				break;
			case n_pointer_member_access:
				{
					CsPointerMemberAccess * node = static_cast<CsPointerMemberAccess*>(*it);
					addTokenToReplace ( getFile ( node ), node->member_token );
				}
				break;
			case n_embedded_type_member_access:
				{
					CsEmbeddedTypeMemberAccess * node = static_cast<CsEmbeddedTypeMemberAccess*>(*it);
					addTokenToReplace ( getFile ( node ), node->member_token );
				}
				break;
			default:
				ASSERT(0);
		}
	}

	{
		std::set<CsEntity*>::iterator it = property_set_.begin(); 
		std::set<CsEntity*>::iterator it_end = property_set_.end(); 
		for ( ; it != it_end; ++it )
		{
			ASSERT ( (*it)->e == ent_property );
			CsEntityProperty * entity = static_cast<CsEntityProperty*>(*it);			CsNode * node = entity->decl;
			switch ( node->type )
			{
				case n_property_declaration:
					{
						CsProperty * decl = static_cast<CsProperty*>(node);
						addTokenToReplace ( getFile ( decl ), decl->compound_identifier.back().identifier_token );
					}
					break;
				default:
					ASSERT(0);
			}
		}
	}

	bNeedGlobalEntityReparsing_ = true;
	replaceTokens ( true );
}

//------------------------------------------------------------------------------

void RefactoringRename::collectTypeProperties	(
	CsEntityType * type,
	CsEntityType * from_type,
	CsEntityProperty * from_property,
	bool bFromDerived
)
{
	//type->traceScopeAndName();
	//property_method->traceScopeAndName();
	CsEntityType * new_from_type = type;
	CsEntityProperty * new_from_property = from_property;
	if ( bFromDerived )
	{
		if ( from_type )
			type_lookup_map_[type].setWasNotifiedFromDerived ( true );
	}
	else
	{
		if ( from_type )
		{
			if ( from_type->e == ent_interface )
				type_lookup_map_[type].setWasNotifiedFromBaseInterface ( true );
			else
				type_lookup_map_[type].setWasNotifiedFromBaseClass ( true );
		}
	}
	bool bStopBaseLookupInClass = false;
	bool bStopBaseLookupInInterfaces = false;
	bool bStopDerivedLookup = false;
	bool bBaseLookupInClass = false;
	bool bBaseLookupInInterfaces = false;
	bool bDerivedLookup = false;
	{
		const ScopeHashTable::hash_id_struct2 * p_his
			= type->scope_table.get_id_list ( from_property->name );
		while ( p_his )
		{
			if ( p_his->data_->e == ent_property
				&& static_cast<CsEntityProperty*>(p_his->data_)->specifier.equal ( from_property->specifier ) )
			{
				CsEntityProperty * property = static_cast<CsEntityProperty*>(p_his->data_);
				//type_lookup_map_[type].method = method;
				if ( property->assembly )
					bAssemblyProperty_ = true;
				if ( bFromDerived )
				{
					if ( property->isOverrideMethod() )
					{
						property_set_.insert ( property );
						bDerivedLookup = true;
					}
					else if ( property->isVirtualMethod() )
					{
						property_set_.insert ( property );
						if ( type->e != ent_interface )
							bStopBaseLookupInClass = true;
						bDerivedLookup = true;
						new_from_property = property;
						new_from_type = type;
					}
					else
					{
						bStopBaseLookupInClass = true;
						bStopBaseLookupInInterfaces = true;
					}
				}
				else
				{
					if ( property->isOverrideMethod() )
					{
						property_set_.insert ( property );
						bBaseLookupInClass = true;
						bBaseLookupInInterfaces = true;
					}
					else if ( property->isVirtualMethod() )
					{
						if ( !from_type
							|| from_type->e == ent_interface )
						{
							property_set_.insert ( property );
						}
						if ( type->e != ent_interface
							&& from_property->parent->e != ent_interface )
						{
							bStopDerivedLookup = true;
						}
						bBaseLookupInClass = true;
						bBaseLookupInInterfaces = true;
						new_from_property = property;
						new_from_type = type;
					}
					else
					{
						bStopDerivedLookup = true;
					}
				}
			}
			p_his = p_his->next_;
		}
	}

	// interface mapping
	if ( from_type
		&& from_type->e != ent_interface )
	{
		bBaseLookupInInterfaces = true;
	}
	if ( from_type
		&& from_type->e == ent_interface )
	{
		bBaseLookupInClass = true;
		bBaseLookupInInterfaces = true;
	}

	if ( !type_lookup_map_[type].isBaseWasLookup() )
	{
		if ( bFromDerived && !bStopBaseLookupInClass
			|| bBaseLookupInClass )
		{
			type_lookup_map_[type].setBaseWasLookup ( true );
			CsEntityClass * base = type->base;
			if ( base )
			{
				if ( !type_lookup_map_[base].isWasNotifiedFromDerived() )
				{
					collectTypeProperties
						( base, new_from_type, new_from_property, true );
				}
			}
		}
		if ( bFromDerived && !bStopBaseLookupInInterfaces
			|| bBaseLookupInInterfaces )
		{
			type_lookup_map_[type].setBaseWasLookup ( true );
			switch ( type->e )
			{
				case ent_class:
				case ent_struct:
				case ent_interface:
					{
						CsEntityTypeWithInterfaces * type_wi
							= static_cast<CsEntityTypeWithInterfaces*>(type);
						if ( type_wi->interfaces )
						{
							single_list1<CsEntityInterface*>::node * next
								= & type_wi->interfaces->first_;
							while ( next )
							{
								CsEntityInterface * base = next->value_;
								if ( base )
								{
									if ( !type_lookup_map_[base].isWasNotifiedFromDerived() )
									{
										collectTypeProperties
											( base, new_from_type, new_from_property, true );
									}
								}
								next = next->next_;
							}
						}
					}
					break;
			}
		}
	}

	if ( !type_lookup_map_[type].isDerivedWasLookup() )
	{
		if ( !bFromDerived && !bStopDerivedLookup
			|| bDerivedLookup )
		{
			type_lookup_map_[type].setDerivedWasLookup ( true );
			if ( type->derived_types )
			{
				single_list1<CsEntityType*>::node * next
					= & type->derived_types->first_;
				while ( next )
				{
					CsEntityType * derived = next->value_;
					if ( type->e == ent_interface )
					{
						if ( !type_lookup_map_[derived].isWasNotifiedFromBaseInterface() )
						{
							collectTypeProperties
								( derived, new_from_type, new_from_property, false );
						}
					}
					else
					{
						if ( !type_lookup_map_[derived].isWasNotifiedFromBaseClass() )
						{
							collectTypeProperties
								( derived, new_from_type, new_from_property, false );
						}
					}
					next = next->next_;
				}
			}
			//if ( type->derived_interfaces )
			//{
			//	single_list1<CsEntityInterface*>::node * next
			//		= & type->derived_interfaces->first;
			//	while ( next )
			//	{
			//		CsEntityInterface * derived = next->value_;
			//		if ( type->e == ent_interface )
			//		{
			//			if ( !type_lookup_map_[derived].isWasNotifiedFromBaseInterface() )
			//			{
			//				collectTypeProperties
			//					( derived, new_from_type, new_from_property, false );
			//			}
			//		}
			//		else
			//		{
			//			if ( !type_lookup_map_[derived].isWasNotifiedFromBaseClass() )
			//			{
			//				collectTypeProperties
			//					( derived, new_from_type, new_from_property, false );
			//			}
			//		}
			//		next = next->next_;
			//	}
			//}
		}
	}
}

//------------------------------------------------------------------------------

void RefactoringRename::renameEvent ( CsEntity * entity )
{
	CsNodeCollectorVisitor visitor;
	visitor.collect_node_flags_[n_simple_name] = true;
	visitor.collect_node_flags_[n_base_member_access] = true;
	visitor.collect_node_flags_[n_member_access] = true;
	visitor.collect_node_flags_[n_pointer_member_access] = true;
	visitor.collect_node_flags_[n_embedded_type_member_access] = true;
	visitor.collect_node_flags_[n_event_declarator] = true;
	visitor.bCheckEntities_ = true;
	visitor.entities_.insert ( entity );

	std::map<std::string,ICsFile*>::iterator fit = files_.begin();
	std::map<std::string,ICsFile*>::iterator fit_end = files_.end();
	for ( ; fit != fit_end; ++fit )
	{
		visitor.collectNodes ( fit->second->getCompilationUnit() );
	}
	MediumFastDeque<CsNode*>::iterator it = visitor.nodes_.begin();
	MediumFastDeque<CsNode*>::iterator it_end = visitor.nodes_.end();
	for ( ; it != it_end; ++it )
	{
		switch ( (*it)->type )
		{
			case n_simple_name:
				{
					CsSimpleName * node = static_cast<CsSimpleName*>(*it);
					addTokenToReplace ( getFile ( node ), node->identifier_token );
				}
				break;
			case n_base_member_access:
				{
					CsBaseMemberAccess * node = static_cast<CsBaseMemberAccess*>(*it);
					addTokenToReplace ( getFile ( node ), node->member_token );
				}
				break;
			case n_member_access:
				{
					CsMemberAccess * node = static_cast<CsMemberAccess*>(*it);
					addTokenToReplace ( getFile ( node ), node->member_token );
				}
				break;
			case n_pointer_member_access:
				{
					CsPointerMemberAccess * node = static_cast<CsPointerMemberAccess*>(*it);
					addTokenToReplace ( getFile ( node ), node->member_token );
				}
				break;
			case n_embedded_type_member_access:
				{
					CsEmbeddedTypeMemberAccess * node = static_cast<CsEmbeddedTypeMemberAccess*>(*it);
					addTokenToReplace ( getFile ( node ), node->member_token );
				}
				break;
			case n_event_declarator:
				{
					CsEventDeclarator * node = static_cast<CsEventDeclarator*>(*it);
					addTokenToReplace ( getFile ( node ), node->compound_identifier.back().identifier_token );
				}
				break;
			default:
				ASSERT(0);
		}
	}

	bNeedGlobalEntityReparsing_ = true;
	replaceTokens ( true );
}

//------------------------------------------------------------------------------

void RefactoringRename::renameType ( CsEntityType * type )
{
	CsNodeCollectorVisitor visitor;
	visitor.collect_node_flags_[n_type_ref] = true;
	visitor.collect_node_flags_[n_using_alias_directive] = true;
	visitor.collect_node_flags_[n_using_namespace_directive] = true;

	visitor.collect_node_flags_[n_class] = true;
	visitor.collect_node_flags_[n_struct] = true;
	visitor.collect_node_flags_[n_interface] = true;
	visitor.collect_node_flags_[n_enum] = true;
	visitor.collect_node_flags_[n_delegate] = true;

	//visitor.collect_node_flags_[n_method_declaration] = true;
	//visitor.collect_node_flags_[n_property_declaration] = true;
	//visitor.collect_node_flags_[n_event_declarator] = true;
	//visitor.collect_node_flags_[n_indexer_declaration] = true;
	
	visitor.bCheckEntities_ = true;
	visitor.bCheckCompoundIdentifier_ = true;
	visitor.entities_.insert ( type );

	std::map<std::string,ICsFile*>::iterator fit = files_.begin();
	std::map<std::string,ICsFile*>::iterator fit_end = files_.end();
	for ( ; fit != fit_end; ++fit )
	{
		visitor.collectNodes ( fit->second->getCompilationUnit() );
	}
	MediumFastDeque<CsNode*>::iterator it = visitor.nodes_.begin();
	MediumFastDeque<CsNode*>::iterator it_end = visitor.nodes_.end();
	for ( ; it != it_end; ++it )
	{
		switch ( (*it)->type )
		{
			case n_type_ref:
				{
					CsTypeRef * typeref
						= static_cast<CsTypeRef*>(*it);
					if ( !typeref->using_directive_used_in_type_lookup ) // ignore typerefs accessed through using directive
					{
						CsEntityTypeRef * entity_typeref = & typeref->entity_typeref;
loop:
						switch ( entity_typeref->type )
						{
							case et_class:
							case et_valuetype:
								{
									CsEntity * entity = entity_typeref->u.entity;
									CsTokenIndex token_index = evalCsTokenIndex
										( typeref->compound_identifier, entity, type );
									if ( token_index )
										addTokenToReplace ( getFile ( typeref ), token_index );
									else
										ASSERT(0);
								}
								break;
							case et_array:
								{
									entity_typeref = & entity_typeref->u.array_specifier->type;
									goto loop;
								}
								break;
						}
					}
				}
				break;
			case n_using_alias_directive:
				{
					CsUsingAliasDirective * ud
						= static_cast<CsUsingAliasDirective*>(*it);
					CsEntity * entity = ud->namespace_or_type_entity;
					if ( entity )
					{
						CsTokenIndex token_index = evalCsTokenIndex
							( ud->namespace_or_type_name, entity, type );
						if ( token_index )
							addTokenToReplace ( getFile ( ud ), token_index );
						else
							ASSERT(0);
					}
				}
				break;
			case n_class:
				{
					CsClass * node = static_cast<CsClass*>(*it);
					addTokenToReplace ( getFile ( node ), node->identifier_token );
				}
				break;
			case n_struct:
				{
					CsStruct * node = static_cast<CsStruct*>(*it);
					addTokenToReplace ( getFile ( node ), node->identifier_token );
				}
				break;
			case n_interface:
				{
					CsInterface * node = static_cast<CsInterface*>(*it);
					addTokenToReplace ( getFile ( node ), node->identifier_token );
				}
				break;
			case n_enum:
				{
					CsEnum * node = static_cast<CsEnum*>(*it);
					addTokenToReplace ( getFile ( node ), node->identifier_token );
				}
				break;
			case n_delegate:
				{
					CsDelegate * node = static_cast<CsDelegate*>(*it);
					addTokenToReplace ( getFile ( node ), node->identifier_token );
				}
				break;
			default:
				ASSERT(0);
		}
	}

	bNeedGlobalEntityReparsing_ = true;
	replaceTokens ( true );
}

//------------------------------------------------------------------------------

CsTokenIndex RefactoringRename::evalCsTokenIndex (
	CsCompoundIdentifier & compound_identifier,
	CsEntity * entity,
	CsEntity * search_entity
)
{
	unsigned n_identifiers = compound_identifier.eval_size();
	while ( n_identifiers && entity )
	{
		if ( entity == search_entity )
			break;
		--n_identifiers;
		entity = entity->parent;
	}
	if ( entity == search_entity )
	{
		--n_identifiers;
		CsCompoundIdentifier::node * next
			= & compound_identifier.first_;
		while ( n_identifiers )
		{
			--n_identifiers;
			next = next->next_;
		}
		return next->value_.identifier_token;
	}
	return 0;
}

//------------------------------------------------------------------------------

void RefactoringRename::renameUsingDirectiveAlias ( CsUsingAliasDirective * using_alias_directive )
{
	CsNodeCollectorVisitor visitor;
	visitor.collect_node_flags_[n_type_ref] = true;
	visitor.collect_node_flags_[n_using_alias_directive] = true;
	visitor.collect_node_flags_[n_using_namespace_directive] = true;
	visitor.bCheckEntities_ = true;
	visitor.bCheckUsingAliasDirective_ = true;
	visitor.using_alias_directive_ = using_alias_directive;
	visitor.collectNodes ( using_alias_directive->parent );

	MediumFastDeque<CsNode*>::iterator it = visitor.nodes_.begin();
	MediumFastDeque<CsNode*>::iterator it_end = visitor.nodes_.end();
	for ( ; it != it_end; ++it )
	{
		switch ( (*it)->type )
		{
			case n_type_ref:
				{
					CsTypeRef * typeref = static_cast<CsTypeRef*>(*it);
					addTokenToReplace ( getFile ( typeref ), typeref->compound_identifier.first_.value_.identifier_token );
				}
				break;
			case n_using_alias_directive:
				{
					CsUsingAliasDirective * ud = static_cast<CsUsingAliasDirective*>(*it);
					if ( ud == using_alias_directive )
						addTokenToReplace ( getFile ( using_alias_directive ), using_alias_directive->identifier_token );
					else
						addTokenToReplace ( getFile ( ud ), ud->namespace_or_type_name.first_.value_.identifier_token );
				}
				break;
			case n_using_namespace_directive:
				{
					CsUsingNamespaceDirective * ud = static_cast<CsUsingNamespaceDirective*>(*it);
					addTokenToReplace ( getFile ( ud ), ud->namespace_or_type_name.first_.value_.identifier_token );
				}
				break;
			default:
				ASSERT(0);
		}
	}

	bNeedGlobalEntityReparsing_ = true;
	replaceTokens ( true );
}

//------------------------------------------------------------------------------

void RefactoringRename::renameNamespace ( CsEntityNamespace * namespace_entity )
{
	CsNodeCollectorVisitor visitor;
	visitor.collect_node_flags_[n_namespace] = true;
	visitor.collect_node_flags_[n_type_ref] = true;
	visitor.collect_node_flags_[n_using_alias_directive] = true;
	visitor.collect_node_flags_[n_using_namespace_directive] = true;

	//visitor.collect_node_flags_[n_method_declaration] = true;
	//visitor.collect_node_flags_[n_property_declaration] = true;
	//visitor.collect_node_flags_[n_event_declarator] = true;
	//visitor.collect_node_flags_[n_indexer_declaration] = true;

	visitor.bCheckEntities_ = true;
	visitor.bCheckCompoundIdentifier_ = true;
	visitor.entities_.insert ( namespace_entity );

	std::map<std::string,ICsFile*>::iterator fit = files_.begin();
	std::map<std::string,ICsFile*>::iterator fit_end = files_.end();
	for ( ; fit != fit_end; ++fit )
	{
		visitor.collectNodes ( fit->second->getCompilationUnit() );
	}
	MediumFastDeque<CsNode*>::iterator it = visitor.nodes_.begin();
	MediumFastDeque<CsNode*>::iterator it_end = visitor.nodes_.end();
	for ( ; it != it_end; ++it )
	{
		switch ( (*it)->type )
		{
			case n_type_ref:
				{
					CsTypeRef * typeref
						= static_cast<CsTypeRef*>(*it);
					if ( !typeref->using_directive_used_in_type_lookup ) // ignore typerefs accessed through using directive
					{
						CsEntityTypeRef * entity_typeref = & typeref->entity_typeref;
loop:
						switch ( entity_typeref->type )
						{
							case et_class:
							case et_valuetype:
								{
									CsEntity * entity = entity_typeref->u.entity;
									CsTokenIndex token_index = evalCsTokenIndex
										( typeref->compound_identifier, entity, namespace_entity );
									if ( token_index )
										addTokenToReplace ( getFile ( typeref ), token_index );
									else
										ASSERT(0);
								}
								break;
							case et_array:
								{
									entity_typeref = & entity_typeref->u.array_specifier->type;
									goto loop;
								}
								break;
						}
					}
				}
				break;
			case n_using_alias_directive:
				{
					CsUsingAliasDirective * ud
						= static_cast<CsUsingAliasDirective*>(*it);
					CsEntity * entity = ud->namespace_or_type_entity;
					if ( entity )
					{
						CsTokenIndex token_index = evalCsTokenIndex
							( ud->namespace_or_type_name, entity, namespace_entity );
						if ( token_index )
							addTokenToReplace ( getFile ( ud ), token_index );
						else
							ASSERT(0);
					}
				}
				break;
			case n_using_namespace_directive:
				{
					CsUsingNamespaceDirective * ud
						= static_cast<CsUsingNamespaceDirective*>(*it);
					CsEntity * entity = ud->namespace_or_type_entity;
					if ( entity )
					{
						CsTokenIndex token_index = evalCsTokenIndex
							( ud->namespace_or_type_name, entity, namespace_entity );
						if ( token_index )
							addTokenToReplace ( getFile ( ud ), token_index );
						else
							ASSERT(0);
					}
				}
				break;
			case n_namespace:
				{
					CsNamespace * decl
						= static_cast<CsNamespace*>(*it);
					CsEntity * entity = decl->entity;
					if ( entity )
					{
						CsTokenIndex token_index = evalCsTokenIndex
							( decl->compound_identifier, entity, namespace_entity );
						if ( token_index )
							addTokenToReplace ( getFile ( decl ), token_index );
						else
							ASSERT(0);
					}
				}
				break;
			default:
				ASSERT(0);
		}
	}

	bNeedGlobalEntityReparsing_ = true;
	replaceTokens ( true );
}

//------------------------------------------------------------------------------

void RefactoringRename::renameLabel ( CsLabel * label )
{
	CsNodeCollectorVisitor visitor;
	visitor.collect_node_flags_[n_goto_statement] = true;

	CsNode * method_node = getMethod ( label );
	if ( !method_node )
	{
		ASSERT(0);
		return;
	}

	visitor.collectNodes ( method_node );
	MediumFastDeque<CsNode*>::iterator it = visitor.nodes_.begin();
	MediumFastDeque<CsNode*>::iterator it_end = visitor.nodes_.end();
	for ( ; it != it_end; ++it )
	{
		switch ( (*it)->type )
		{
			case n_goto_statement:
				{
					CsGotoStatement * stat
						= static_cast<CsGotoStatement*>(*it);
					if ( stat->label_type == tkIDENTIFIER
						&& stat->goto_target == label )
					{
						addTokenToReplace ( getFile ( stat ), stat->label_token );
					}
				}
				break;
			default:
				ASSERT(0);
		}
	}
	if ( label )
		addTokenToReplace ( getFile ( label ), label->identifier_token );

	replaceTokens ( true );
}

//------------------------------------------------------------------------------

void RefactoringRename::replaceTokens ( bool bLocalRename )
{
	token_to_rename_iterator it = tokensToReplace_.begin();
	token_to_rename_iterator it_end = tokensToReplace_.end();
	for ( ; it != it_end; ++it )
	{
		ICsFile * file = it->first;
		CsTokenIndex token_index = it->second;

		MediumFastDeque<CsToken>::iterator it = file->getTokens().begin();
		std::advance ( it, token_index - 1 );
		ASSERT ( it->type == tkIDENTIFIER );
		it->text = new_name_;
	}
	tokensToReplace_.clear();
}

//------------------------------------------------------------------------------

ICsFile * RefactoringRename::findFile ( const char * path )
{
	std::map<std::string,ICsFile*>::iterator it = files_.find ( path );
	if ( it != files_.end() )
		return it->second;
	for ( it = files_.begin(); it != files_.end(); ++it )
	{
		std::string file_name;
		GetFileFromPath ( it->first.c_str(), file_name );
		if ( file_name == path )
			return it->second;
	}
	return 0;
}

//------------------------------------------------------------------------------

CsFile * RefactoringRename::getFile ( CsNode * node )
{
	while ( node )
	{
		if ( node->type == n_compilation_unit )
		{
			return static_cast<CsCompilationUnit*>(node)->file;
		}
		node = node->parent;
	}
	ASSERT(0);
	return 0;
}

//------------------------------------------------------------------------------

CsNode * RefactoringRename::getMethod ( CsNode * node )
{
	while ( node )
	{
		switch ( node->type )
		{
			case n_constructor_declaration:
			case n_method_declaration:
			case n_property_accessor:
			case n_event_accessor:
			case n_operator_declaration:
			case n_conversion_operator_declaration:
			case n_destructor_declaration:
				{
					return node;
				}
				break;
		}
		node = node->parent;
	}
	ASSERT(0);
	return 0;
}

//------------------------------------------------------------------------------

MediumFastDeque<CsToken>::iterator RefactoringRename::getToken ( ICsFile * file, unsigned begin, unsigned end )
{
	MediumFastDeque<CsToken>::iterator it = file->getTokens().begin();
	MediumFastDeque<CsToken>::iterator it_end = file->getTokens().end();
	unsigned offset = 0;
	for ( ; it != it_end; ++it )
	{
		offset += it->evalSize();
		if ( offset < end )
			break;
	}
	return it;
}

//------------------------------------------------------------------------------
// CsFindTokenVisitor
//------------------------------------------------------------------------------

CsNode * CsFindTokenVisitor::findNodeByTokenIndex
	( CsNode * startNode, CsTokenIndex index )
{
	tmp_found_node = 0;
	tmp_token_index = index;
	startNode->visit ( this );
	return tmp_found_node;
}

//------------------------------------------------------------------------------

void CsFindTokenVisitor::visitNode ( CsNode * node )
{
	tmp_tokens.resize0();
	node->addTokens ( tmp_tokens );
	MediumFastDeque<CsTokenIndex>::iterator it = tmp_tokens.tokens.begin();
	MediumFastDeque<CsTokenIndex>::iterator it_end = tmp_tokens.tokens.end();
	for ( ; it != it_end; ++it )
	{
		if ( *it == tmp_token_index )
			tmp_found_node = node;
	}
}

//------------------------------------------------------------------------------
// CsNodeCollectorVisitor
//------------------------------------------------------------------------------

CsNodeCollectorVisitor::CsNodeCollectorVisitor()
{
	init();
}

//------------------------------------------------------------------------------

void CsNodeCollectorVisitor::init()
{
	nodes_.clear();
	for ( int i = 0; i != n_end; ++i )
	{
		collect_node_flags_[i] = false;
	}
	bCheckEntities_ = false;
	bCheckCompoundIdentifier_ = false;
	bCheckUsingAliasDirective_ = false;
	using_alias_directive_ = 0;
	bProjectEntities_ = false;
}

//------------------------------------------------------------------------------

void CsNodeCollectorVisitor::collectNodes ( CsNode * startNode )
{
	startNode->visit ( this );
}

//------------------------------------------------------------------------------

void CsNodeCollectorVisitor::visitNode ( CsNode * node )
{
	if ( collect_node_flags_[node->type] )
	{
		if ( bCheckUsingAliasDirective_ )
		{
			switch ( node->type )
			{
				case n_type_ref:
					{
						CsTypeRef * typeref = static_cast<CsTypeRef*>(node);
						if ( typeref->using_directive_used_in_type_lookup == using_alias_directive_ )
							nodes_.push_back ( typeref );
					}
					break;
				case n_using_alias_directive:
					{
						CsUsingAliasDirective * ud = static_cast<CsUsingAliasDirective*>(node);
						if ( ud == using_alias_directive_
							|| ud->using_directive_used_in_type_lookup == using_alias_directive_ )
						{
							nodes_.push_back ( ud );
						}
					}
					break;
				case n_using_namespace_directive:
					{
						CsUsingNamespaceDirective * ud = static_cast<CsUsingNamespaceDirective*>(node);
						if ( ud->using_directive_used_in_type_lookup == using_alias_directive_ )
						{
							nodes_.push_back ( ud );
						}
					}
					break;
				default:
					ASSERT(0);
			}
		}
		else
		{
			if ( bCheckEntities_ )
			{
				bool bCheckOther = false;
				if ( bCheckCompoundIdentifier_ )
				{
					if ( node->type == n_type_ref )
					{
						CsTypeRef * typeref = static_cast<CsTypeRef*>(node);
						CsEntityTypeRef * entity_typeref = & typeref->entity_typeref;
loop:
						switch ( entity_typeref->type )
						{
							case et_class:
							case et_valuetype:
								{
									CsEntity * entity = entity_typeref->u.entity;
									if ( isEntityExist ( typeref->compound_identifier, entity, typeref->using_directive_used_in_type_lookup != 0 ) )
										nodes_.push_back ( typeref );
								}
								break;
							case et_array:
								{
									entity_typeref = & entity_typeref->u.array_specifier->type;
									goto loop;
								}
								break;
						}
					}
					else if ( node->type == n_using_alias_directive )
					{
						CsUsingAliasDirective * ud = static_cast<CsUsingAliasDirective*>(node);
						if ( ud->namespace_or_type_entity
							/*&& ud->entity->e != ent_namespace*/ )
						{
							if ( isEntityExist ( ud->namespace_or_type_name, ud->namespace_or_type_entity, ud->using_directive_used_in_type_lookup != 0 ) )
								nodes_.push_back ( ud );
						}
					}
					else if ( node->type == n_using_namespace_directive )
					{
						CsUsingNamespaceDirective * ud = static_cast<CsUsingNamespaceDirective*>(node);
						if ( ud->namespace_or_type_entity
							/*&& ud->entity->e != ent_namespace*/ )
						{
							if ( isEntityExist ( ud->namespace_or_type_name, ud->namespace_or_type_entity, ud->using_directive_used_in_type_lookup != 0 ) )
								nodes_.push_back ( ud );
						}
					}
					//else if ( node->type == n_using_directive )
					//{
					//	CsUsingDirective * using_directive = static_cast<CsUsingDirective*>(node);
					//	if ( using_directive->bUsingAliasDirective
					//		&& using_directive->namespace_or_type_entity
					//		/*&& using_directive->entity->e != ent_namespace*/ )
					//	{
					//		if ( isEntityExist ( using_directive->namespace_or_type_name, using_directive->namespace_or_type_entity ) )
					//			nodes_.push_back ( using_directive );
					//	}
					//}
					else if ( node->type == n_namespace )
					{
						CsNamespace * decl = static_cast<CsNamespace*>(node);
						//decl->traceLine();
						if ( decl->entity )
						{
							if ( isEntityExist ( decl->compound_identifier, decl->entity ) )
								nodes_.push_back ( decl );
						}
					}
					else
					{
						bCheckOther = true;
					}
				}
				else
				{
					bCheckOther = true;
				}
				if ( bCheckOther )
				{
					CsEntityCollector entityCollector;
					node->addEntityDefinitions ( entityCollector );
					node->addEntityReferences ( entityCollector );
					MediumFastDeque<CsEntity*>::iterator it = entityCollector.entities.begin();
					MediumFastDeque<CsEntity*>::iterator it_end = entityCollector.entities.end();
					for ( ; it != it_end; ++it )
					{
						if ( entities_.find ( *it ) != entities_.end() )
							nodes_.push_back ( node );
					}
				}
			}
			else
			{
				nodes_.push_back ( node );
			}
		}
	}
}

//------------------------------------------------------------------------------

bool CsNodeCollectorVisitor::isEntityExist (
	CsCompoundIdentifier & compound_identifier,
	CsEntity * entity,
	bool bSkipFirstEntity
)
{
	unsigned n_identifiers = compound_identifier.eval_size();
	while ( n_identifiers && entity )
	{
		if ( bSkipFirstEntity )
			bSkipFirstEntity = !bSkipFirstEntity;
		else
		{
			if ( entities_.find ( entity ) != entities_.end() )
				return true;
		}
		--n_identifiers;
		entity = entity->parent;
	}
	return false;
}

//------------------------------------------------------------------------------
// CsEntityCollectorVisitor
//------------------------------------------------------------------------------

CsEntityCollectorVisitor::CsEntityCollectorVisitor()
{
	init();
}

//------------------------------------------------------------------------------

void CsEntityCollectorVisitor::init()
{
	entities.clear();
	for ( int i = 0; i != ent_end; ++i )
	{
		collect_entities_flags[i] = false;
	}
	bCollectDerivedTypes = false;
	base_type = 0;
}

//------------------------------------------------------------------------------

void CsEntityCollectorVisitor::collectEntities ( CsEntity * startEntity )
{
	startEntity->visit ( this );
}

//------------------------------------------------------------------------------

void CsEntityCollectorVisitor::visitEntity ( CsEntity * entity )
{
	if ( collect_entities_flags[entity->e] )
	{
		if ( bCollectDerivedTypes )
		{
			switch ( entity->e )
			{
				case ent_class:
				case ent_struct:
				case ent_interface:
				case ent_enum:
				case ent_delegate:
					{
						if ( CsEntityType::isBaseDerived ( base_type, static_cast<CsEntityType*>(entity) ) )
							entities.push_back ( entity );
					}
					break;
			}
		}
		else
		{
			entities.push_back ( entity );
		}
	}
}

//------------------------------------------------------------------------------

